/****************************************************************************
*
*	  Module Filename:  Realisable_Interaction_Type.cpp
*	  Module Name:      Realisable_Interaction_Type
*	  Description:      An object of this class represents an Interaction_Type
*                       which is currently realisable in the Reactor.
*                       That is, if a particular Realisable_Interaction_Type exists,
*                       the Reactant(s) and any Catalyst required
*                       for the Interaction to occur are currently present
*                       in the Reactor.
*	  Author:   		Chris Gordon-Smith
*
*     Copyright (C) 2003, 2004, 2005, 2006, 2008 Chris Gordon-Smith
*     Email: See http://www.simsoup.info/
*
*     This program is free software; you can redistribute it and/or modify it
*     under the terms of the GNU General Public License Version 2, as published 
*     by the Free Software Foundation.
*
*     This program is distributed in the hope that it will be useful, but
*     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
*     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
*     for more details.
*
*     You should have received a copy of the GNU General Public License along
*     with this program; if not, write to the Free Software Foundation, Inc.,
*     59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
*     See the file Licence.txt for a copy of the GNU General Public License.
*
*	  Amendment History:-
*
*		Version	   Date	   			Update Description
*		=======    ===========		==================
*
*
*
*
****************************************************************************/

#include    "globldef.h"
#include    "Library_Headers.h"

namespace DN {
#include "list.hpp"        // Special purpose - Non STL
}

//--------------  Reactor  Subsystem  -------------------------------------------------------------------------
#include    "Realisable_Interaction_Type.h"
#include    "Realised_Interaction_Type.h"
#include    "Realised_Molecule_Type.h"
#include    "Reactor.h"
//---------------------------------------------------------------------------------------------------------------

//--------------  Chemistry  Subsystem  -------------------------------------------------------------------------
#include    "Interaction_Type.h"
#include    "Molecule_Type.h"
//---------------------------------------------------------------------------------------------------------------

//----------------------------------------------------------------------------------------------------------
DN::Node<Realisable_Interaction_Type*>*  Realisable_Interaction_Type::Get_Node_In_Reactors_List_Ptr() const

//  Get a pointer to the list node in the Reactor's list of Realisable_Interaction_Types
//  that identifies this object

{
    return  Node__In_Reactors_List_Ptr;
}
//----------------------------------------------------------------------------------------------------------
void  Realisable_Interaction_Type::Set_Node_In_Reactors_List_Ptr(DN::Node<Realisable_Interaction_Type*>*  Node__Ptr)

//  Set the pointer to the list node in the Reactor's list of Realisable_Interaction_Types
//  that identifies this object

{
   Node__In_Reactors_List_Ptr = Node__Ptr;
}
//----------------------------------------------------------------------------------------------------------
bool  Realisable_Interaction_Type::Get_Unrealisable() const

//  Return the Unrealisable indicator

{
    return(Unrealisable);
}
//----------------------------------------------------------------------------------------------------------------

void Realisable_Interaction_Type::Set_Unrealisable()

//  Set the Unrealisable indicator and also unlink from the Interaction_Type

{   
    Unrealisable = true;
    Interaction_Type__Ptr->Unlink_From_R_Int_Type(this);
}
//----------------------------------------------------------------------------------------------------------
double Realisable_Interaction_Type::Get_Interaction_Rate() const

//  Return the Realisable_Interaction_Type's Interaction_Rate

{
    return(Interaction_Rate);
}
//----------------------------------------------------------------------------------------------------------
Interaction_Type*  const  Realisable_Interaction_Type::Get_Interaction_Type_Ptr() const

//  Return a pointer to the Interaction_Type that corresponds to
//  the Realisable_UInteraction_Type

{
    return(Interaction_Type__Ptr);
}
//----------------------------------------------------------------------------------------------------------
Realisable_Interaction_Type::Realisable_Interaction_Type(Interaction_Type*  Int_Type_Ptr)

//  Constructor  - Link to the associated Interaction_Type (both ways) and initialise Interaction_Rate
//  Also create a Realised Interaction_Type if needed, and link the Realisable_Interaction_Type to it

{
    Int_Type_Ptr->Link_To_R_Int_Type(this);
    Interaction_Type__Ptr =  Int_Type_Ptr;
    Interaction_Rate = 0;
    Unrealisable = false;
    Enabled = true;
    
//  Create a Realised_Interaction_Type object if the Interaction_Type is not already realised

    if (not Int_Type_Ptr->Get_Realised_Int_Type_Ptr())
    THEN     
        Realised_Interaction_Type* Realised_Int_Type_Ptr;
        Realised_Int_Type_Ptr = Reactor::Get_Reactor__Ptr()->Add_Realised_Int_Type(Int_Type_Ptr);
        Int_Type_Ptr->Link_To_Realised_Int_Type(Realised_Int_Type_Ptr);  
    ENDIF       
}

//------------------------------------------------------------------------------

double Realisable_Interaction_Type::Update_Interaction_Rates(int Reactor_Volume) 

//  Setup arguments and use the Interaction_Type (derived) object to Calculate the 
//  Interaction_Rate for the Realisable_Interaction_Type based on the Reactor contents, 
//  using the specific rules for the derived class.

//  Set the Realisable_Interaction_Type's Interaction_Rate accordingly, and also return the rate
//  to the caller.

//  If the Realisable_Interaction_Type is disabled, then return 0

{
    if (Enabled not_eq true)
    THEN
        return 0.0;
    ENDIF
    
    struct  Calculate_Reactor_Interaction_Rates_Args Args;

//................................................................................................................

//  Setup arguments used for calculating the forward rate

    Args.Reactant1_Num_Mols
      = Get_Interaction_Type_Ptr()->Get_Reactant1_MolType_Ptr()->Get_Realised_Molecule_Type__Ptr()->Get_Num_Of_Mols_In_Reactor();

    Molecule_Type*  MolType_R2_Ptr = Get_Interaction_Type_Ptr()->Get_Reactant2_MolType_Ptr();
    if (MolType_R2_Ptr == 0)
    THEN
        Args.Reactant2_Num_Mols = -1;  // Used to trigger Assertion in the overloaded
                                       // Calculate_Reactor_Interaction_Rate function
    ELSE
        Args.Reactant2_Num_Mols
      = MolType_R2_Ptr->Get_Realised_Molecule_Type__Ptr()->Get_Num_Of_Mols_In_Reactor();
    ENDIF
//................................................................................................................

//  Setup arguments used for calculating the reverse rate
//  It is possible that the Product Molecule Type(s) have never been realised. If so, the 
//  Product Molecule Type pointers in the Interaction_Type will be zero. we therefore
//  test for a zero pointer and set the number of Product Molecules accordingly.

    Realised_Molecule_Type* R_MolType_P1_Ptr;
    R_MolType_P1_Ptr = Get_Interaction_Type_Ptr()->Get_Product1_MolType_Ptr()->Get_Realised_Molecule_Type__Ptr();
    
    if (R_MolType_P1_Ptr == 0)
    THEN
        Args.Product1_Num_Mols = 0;
    ELSE 
        Args.Product1_Num_Mols = R_MolType_P1_Ptr->Get_Num_Of_Mols_In_Reactor();
    ENDIF

    Molecule_Type*  MolType_P2_Ptr = Get_Interaction_Type_Ptr()->Get_Product2_MolType_Ptr();
    if (MolType_P2_Ptr == 0)  // This Interaction Type does not have a Product 2
    THEN
        Args.Product2_Num_Mols = -1;  // Used to trigger Assertion in the overloaded
                                    // Calculate_Reactor_Interaction_Rate function
    ELSE
        Realised_Molecule_Type* R_MolType_P2_Ptr = MolType_P2_Ptr->Get_Realised_Molecule_Type__Ptr();
        if(R_MolType_P2_Ptr == 0)
        THEN
            Args.Product2_Num_Mols = 0;
        ELSE
            Args.Product2_Num_Mols = R_MolType_P2_Ptr->Get_Num_Of_Mols_In_Reactor();
        ENDIF
    ENDIF
//................................................................................................................

    Args.Reactor_Volume = Reactor_Volume;   
    return(Interaction_Rate = Interaction_Type__Ptr->Calculate_Reactor_Interaction_Rate(&Args));
}

