/****************************************************************************
*
*     Header Filename:  Indexed_Series.cpp
*     Module Name:      Indexed_Series
*     Description:      Stores and maintains a series of values that are accessible by an Index
*                       The Index and teh Value together can be used to represent a data pair
*	  Author:           Chris Gordon-Smith
*
*     Copyright (C) 2005, 2006, 2008 Chris Gordon-Smith
*     Email: See http://www.simsoup.info/
*
*     This program is free software; you can redistribute it and/or modify it
*     under the terms of the GNU General Public License Version 2, as published 
*     by the Free Software Foundation.
*
*     This program is distributed in the hope that it will be useful, but
*     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
*     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
*     for more details.
*
*     You should have received a copy of the GNU General Public License along
*     with this program; if not, write to the Free Software Foundation, Inc.,
*     59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
*     See the file Licence.txt for a copy of the GNU General Public License.
*	  Amendment History:-
*
*         Version    Date                 Update Description
*         =======    ===========          ==================
*
*
*
*
****************************************************************************/
//  Comment out to disallow tracing this file
//#define  ENABLE_TRACE_FILE  

#include  "globldef.h"
#include  "Library_Headers.h"

//--------------  Statistics  Subsystem  -----------------------------------------------------------------------------

#include  "Indexed_Series.h"
//---------------------------------------------------------------------------------------------------------------
void  Indexed_Series::Add_Value(const double Index, double Value)

//  Add the specified value to the Series.
//  If the Treat_No_Entry_As_Zero ("Sparse" or emmory saving Mode) flag is set, then don't add anything if
//  Value is zero

{  
    if (not Treat_No_Entry_As_Zero or (Value not_eq 0))
    THEN
        pair<double,double>  P(Index, Value);
        Value_List.insert(P);
    ENDIF
                                      
//  Update the Series_Ranges as required (all range values will be set for first data point)
    
    if (Index < Ranges.Index_Min) Ranges.Index_Min = Index;
    if (Index > Ranges.Index_Max) Ranges.Index_Max = Index;
    if (Value < Ranges.Value_Min) Ranges.Value_Min = Value;
    if (Value > Ranges.Value_Max) Ranges.Value_Max = Value;    
}
//-----------------------------------------------------------------------------------------------------------------

bool  Indexed_Series::Axes_Are_Changed()

//   Check whether the axes for the Indexed Series have changed
//   Return true if they have, otherwise false

{     
    if((Current_Index_Axis.First_Tick not_eq Previous_Index_Axis.First_Tick)
                or
       (Current_Index_Axis.Last_Tick not_eq  Previous_Index_Axis.Last_Tick)
                or
       (Current_Value_Axis.First_Tick not_eq Previous_Value_Axis.First_Tick)
                or
       (Current_Value_Axis.Last_Tick not_eq  Previous_Value_Axis.Last_Tick))       
    THEN
        return true;
    ELSE
        return false;
    ENDIF
}
//-----------------------------------------------------------------------------------------------------------------

bool Indexed_Series::Axis_Is_Changed(double Old_Range_Min, 
                                     double Old_Range_Max, 
                                     double New_Range_Min, 
                                     double New_Range_Max)

//  Given two ranges, determine whether they require different 
//  axes (ie different first and last ticks)

{
    double Old_First_Tick;
    double Old_Last_Tick;    
    double New_First_Tick;
    double New_Last_Tick;    
    double Increment;
    
    Ticks_And_Increment_For_Range(Old_Range_Min, 
                                  Old_Range_Max,    
                                  Old_First_Tick, 
                                  Old_Last_Tick, 
                                  Increment);
    Ticks_And_Increment_For_Range(New_Range_Min, 
                                  New_Range_Max, 
                                  New_First_Tick, 
                                  New_Last_Tick, 
                                  Increment);    
        
    if ((Old_First_Tick not_eq New_First_Tick)
                or
        (Old_Last_Tick not_eq New_Last_Tick))
    THEN
        return true;
    ELSE
        return false;
    ENDIF
}
//--------------------------------------------------------------------------------------------------------------

void  Indexed_Series::Clear()

//   Clear all Data Values in the Series and initialise ranges and axes

{
    Value_List.clear();
    Init_Ranges_And_Axes();
}
//-----------------------------------------------------------------------------------------------------------------

bool   Indexed_Series::Get_Value_For_Index(double& Index, double& Value)

//  Get the Value for the specified Index.  
//  If the Treat_No_Entry_As_Zero ("Sparse" or memory saving Mode) flag is set, 
//  then set Value = zero if no value exists for the Index. 
//  In this mode always return true.
//  
//  If the Treat_No_Entry_As_Zero flag is not set, 
//  then return true if a Value exists, false if not

//  The interface supports non integral Index, but currently a check is made that it must be integral.
//  Non integral Indexes could be problematic for some uses (eg Manhattan Series). The logic
//  below will return zero for a particular index (eg 10.000001) even if ther is a value for index = 10.
//  Hence the insistence that Index must be integral.
{
    int IntInd = static_cast <int> (Index);
    double DblInd = IntInd;
    Assert(Index == DblInd, "Indexed_Series::Get_Value_For_Index - Index is not an integer value");

    if (Value_List.find(Index) not_eq Value_List.end())
    THEN
        Value = Value_List[Index];
        return true;
    ELSEIF (Treat_No_Entry_As_Zero)
    THEN
        Value = 0;
        return true;
    ELSE
        return false;
    ENDIF
}
//---------------------------------------------------------------------------------------------------------------

bool  Indexed_Series::Get_First_Data_Pair(double& Index, double& Value)

//  Return the Index and Value from the first position in the series

{
    if (Value_List.empty())
    THEN
        return false;
    ELSE
        Value_List_Iter = Value_List.begin();
        Index = (*Value_List_Iter).first;
        Value = (*Value_List_Iter).second;
        return true;
    ENDIF
}
//-----------------------------------------------------------------------------------------------------------------

bool  Indexed_Series::Get_Next_Data_Pair(double& Index, double& Value)

//  Increment the iterator to the next Index + Value pair 
//  in the Series and return the pair

{
    if (++Value_List_Iter == Value_List.end())
    THEN
        return false;
    ELSE
        Index = (*Value_List_Iter).first;
        Value = (*Value_List_Iter).second;
        return true;
    ENDIF
}
//----------------------------------------------------------------------------------------------------------------

int   Indexed_Series::Get_Num_Values()

//  Return the number of entries in the Indexed_Series

{
    return  Value_List.size();
}
//---------------------------------------------------------------------------------------------------------------

string  Indexed_Series::Get_Series_List_Name()

//  Return the name of the Series List to which the Indexed Series belongs

{
    TRACE("Indexed_Series::Get_Series_List_Name()");
    return Series_List_Name_;
}

//----------------------------------------------------------------------------------------------------------------
string  Indexed_Series::Get_Indexed_Series_Name() const

//  Return the Series Name

{
    return Indexed_Series_Name_;
}
//---------------------------------------------------------------------------------------------------------------

void  Indexed_Series::Get_Series_Ranges(double& Index_Min,double& Index_Max,
                                  double& Value_Min,double& Value_Max)
                                                 
//  Get the max and min values for the Index and the Value for the Indexed_Series

{
    Index_Min = Ranges.Index_Min;
    Index_Max = Ranges.Index_Max;
    Value_Min = Ranges.Value_Min;
    Value_Max = Ranges.Value_Max;
}
//----------------------------------------------------------------------------------------------------------------

void    Indexed_Series::Get_Index_Axis(double& First_Tick, double& Last_Tick, double& Increment)

//  Pass back the Index Axis values

{
    First_Tick = Current_Index_Axis.First_Tick;
    Last_Tick = Current_Index_Axis.Last_Tick;    
    Increment = Current_Index_Axis.Increment;    
}
//---------------------------------------------------------------------------------------------------------------

void    Indexed_Series::Get_Value_Axis(double& First_Tick, double& Last_Tick, double& Increment)

//  Pass back Value Axis values

{
    First_Tick = Current_Value_Axis.First_Tick;
    Last_Tick = Current_Value_Axis.Last_Tick;    
    Increment = Current_Value_Axis.Increment;    
}
//----------------------------------------------------------------------------------------------------------------

Indexed_Series::Indexed_Series(string Series_Name,   
                               string Series_List_Name,
                               string Index_Name, 
                               string Value_Name, 
                               bool No_Entry_Means_Zero)

//  Constructor

{
    TRACE("Indexed_Series::Indexed_Series");
    
    Indexed_Series_Name_ = Series_Name;
    Series_List_Name_ = Series_List_Name;

    Index_Name_ = Index_Name;
    Value_Name_ = Value_Name;    
    Display_Flag = false;  
        
    Treat_No_Entry_As_Zero = No_Entry_Means_Zero;

    Init_Ranges_And_Axes();
}
//----------------------------------------------------------------------------------------------------------------

void   Indexed_Series::Init_Ranges_And_Axes()

//  Initialise the Series_Ranges. Set Min to MAX and vice versa. This enables Add_Data_Value
//  to check against Min / Max and adjust as necessary.
//  Also initialise the Axis information


{
    Ranges.Index_Min = 0;  // Currently all series are used for time which always starts at 0
    Ranges.Index_Max = -DBL_MAX;  // In general, Index can be negative
    Ranges.Value_Min = DBL_MAX;
    Ranges.Value_Max = -DBL_MAX;    
  
    Previous_Index_Axis.First_Tick = -DBL_MAX;
    Previous_Index_Axis.Last_Tick = DBL_MAX;
    Previous_Index_Axis.Increment = DBL_MAX;
    Previous_Value_Axis.First_Tick = -DBL_MAX;
    Previous_Value_Axis.Last_Tick = DBL_MAX;
    Previous_Value_Axis.Increment = DBL_MAX;
}
//-----------------------------------------------------------------------------------------------------------------

bool  Indexed_Series::Is_For_Display()

//  Return value indicates whether the Series should be displayed

{
    return  Display_Flag;
}
//----------------------------------------------------------------------------------------------------------------

bool  Indexed_Series::No_Entry_Means_Zero()

//  Return whether or not a zero entry means a zero value

{
    return Treat_No_Entry_As_Zero;
}
//----------------------------------------------------------------------------------------------------------------
void   Indexed_Series::Ticks_And_Increment_For_Range(double Range_Min, double Range_Max, 
                                                  double& First_Tick, double& Last_Tick, double& Increment)

//  Determine the appropriate maximum and maximum tick values for the Range. Tick
//  values can be used for the tick points when displaying an axis corresponding to the Range. 
//  Also calculate the increments between ticks. 

//  The increments between ticks should be 1 x 10^n, 2 x 10^n or 5 x 10^n. 
//  This is determined on the basis of the range of values according to the 
//  following rule (where n is any integer):-

//  1 x 10^n <= Range < 2 x 10^n:    Increment = 2 x 10^n-1
//  2 x 10^n <= Range < 5 x 10^n:    Increment = 5 x 10^n-1
//  5 x 10^n <= Range < 1 x 10^n+1:  Increment = 1 x 10^n

//  The minimum and maximum tick are determined according to the following rule:-

//  First_Tick = (integer part of (Range_Min / Increment)) x Increment
//  Last_Tick  = ((integer part of (Range_Max / X_Increment)) x Increment) + X_Increment

{    
// Adjust range values to deal with case where max and min are same
    
    if (Range_Max == Range_Min)
    THEN
        Range_Max++;
        Range_Min--;
    ENDIF  
    
    double Range = Range_Max - Range_Min;  // Range of values
    
    double Range_Exponent = log10(Range);
    
//  If negative exponent, then must still round down (towards more negative).
//  but only if Range exponent is not already an integer
    
    double Dummy;
    if ((Range_Exponent < 0) and (modf(Range_Exponent, &Dummy) not_eq 0.0))
    THEN
        Range_Exponent = Range_Exponent - 1;
    ENDIF

//  Get Standard form Multiplier for Range
        
    double Range_Multiplier = Range / pow(10.0, int(Range_Exponent));

//  Determine increment for ticks
        
    if     (1 <= Range_Multiplier and Range_Multiplier < 2)  
    THEN
        Increment = 2 * (pow(10.0,int(Range_Exponent) - 1)); 
    ELSEIF (2 <= Range_Multiplier and Range_Multiplier < 5)
    THEN
        Increment = 5 * (pow(10.0,int(Range_Exponent) - 1)); 
    ELSEIF (5 <= Range_Multiplier and Range_Multiplier < 10)
    THEN
        Increment = 1 * pow(10.0,int(Range_Exponent));
    ELSE
        Assert (true == false, "Indexed_Series::Ticks_And_Increment_For_Range - Standard Form Multiplier out of range.");

//             - Standard Form Multiplier out of range. Range_Min = ")
//                + string(To_String(Range_Min))
//                + string(" Range_Max = ") + string(To_String(Range_Max))
//                + string(" Range_Multiplier = ") + string(To_String(Range_Multiplier))
//                + string(" Range_Exponent = ") + To_String(Range_Exponent)
//                + string(" int(Range_Exponent) = ") + To_String(int(Range_Exponent))
//                      ).c_str());    
    ENDIF
    
//  Determine first and last tick values for the range

    First_Tick = int(Range_Min / Increment) * Increment;
    Last_Tick =  (int(Range_Max / Increment) * Increment) + Increment;
}   
//-----------------------------------------------------------------------------------------------------------------

void  Indexed_Series::Toggle_Series_Display()

//  Toggle the flag that specifies whether or not the Series should be displayed.

{
    TRACE("Indexed_Series::Toggle_Series_Display");
    Display_Flag = not Display_Flag;
}
//---------------------------------------------------------------------------------------------------------------

void    Indexed_Series::Update_Axes()

//  Update Index and Value axes and remember previous values

{
    Previous_Index_Axis.First_Tick = Current_Index_Axis.First_Tick;
    Previous_Index_Axis.Last_Tick = Current_Index_Axis.Last_Tick;    
    Previous_Index_Axis.Increment = Current_Index_Axis.Increment;    
    
    Previous_Value_Axis.First_Tick = Current_Value_Axis.First_Tick;
    Previous_Value_Axis.Last_Tick = Current_Value_Axis.Last_Tick;    
    Previous_Value_Axis.Increment = Current_Value_Axis.Increment;    

    Ticks_And_Increment_For_Range(Ranges.Index_Min, Ranges.Index_Max, 
                                  Current_Index_Axis.First_Tick, Current_Index_Axis.Last_Tick, Current_Index_Axis.Increment);
    Ticks_And_Increment_For_Range(Ranges.Value_Min, Ranges.Value_Max, 
                                  Current_Value_Axis.First_Tick, Current_Value_Axis.Last_Tick, Current_Value_Axis.Increment);
}
//------------------------------------------------------------------------------------------------------------------        
        
